# IMPORTS
import logging
import pandas as pd
import re

# FUNCTIONS
def split_dataframe(df: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Splits input dataframe into two pandas.DataFrames according to column name.

    :param df: is the input dataframe

    :return: two pandas.DataFrames (donor DataFrame and recipient DataFrame)
    """
    return df.filter(regex='_D').copy(), df.filter(regex='_R').copy()


def _is_eplet_to_be_added(
        eplet: str,
        suffix: str,
        df_data: pd.DataFrame,
        verifiedonly: bool = False
) -> bool:
    """
    Checks whether the eplet should be added
    :param eplet: is it the correct eplet
    :param suffix: is it the correct gene (needed because for example 26L is a DR eplet but also a DQ eplet)
    :param df_data: reference pandas.DataFrame with details about the eplets
    :param verifiedonly: whether or not to take into account only verified eplets

    :return: whether or not the eplet should be added to the list of eplets
    """

    if not verifiedonly:
        # if verified only is False, then we don't care about the 'confirmation' column
        return True

    locus_det: str
    if eplet[0] in ["R", "Q", "P"]:
        locus_det = "i2"
    else:
            locus_det = suffix

    return (
        df_data[
            (df_data["eplet"] == eplet)
            &
            (df_data["locus"] == locus_det)
        ]["confirmation"].item() == "Yes"
    )



def _convert_to_eplets(
        allele: str,
        df_ref: pd.DataFrame,
        suffix: str,
        df_data: pd.DataFrame,
        interlocus2: bool,
        verifiedonly: bool = False
) -> list[str]:
    """
    :param allele: allele to convert
    :param df_ref: reference pandas.DataFrame for the locus of the allele
    :param suffix: which locus the allele belongs to
    :param df_data: reference pandas.DataFrame with details about all the eplets (ep_data.csv)
    :param interlocus2: whether or not to take into account interlocus eplets for HLA of class II
    :param verifiedonly: whether or not to take into account only verified eplets

    :return: list of eplets corresponding to the input allele
    """

    eplet_list: list[str] = []
    for eplet in df_ref.loc[allele].values:
        if not pd.isnull(eplet):
            if _is_eplet_to_be_added(eplet, suffix, df_data, verifiedonly):
                if eplet[0] in ["R", "Q", "P"]:
                    if interlocus2:
                        eplet_list.append(eplet)
                else:
                    eplet_list.append(f"{eplet}_{suffix}")

    return eplet_list



def _allele_df_to_eplets_df(
        df: pd.DataFrame,
        df_a: pd.DataFrame,
        df_b: pd.DataFrame,
        df_c: pd.DataFrame,
        df_dr: pd.DataFrame,
        df_dq: pd.DataFrame,
        df_dp: pd.DataFrame,
        df_data: pd.DataFrame,
        interlocus2: bool,
        verifiedonly: bool = False
) -> pd.DataFrame:
    """
    Transformation of the allele dataframe (index + alleles) into an eplet dataframe (index + eplets)

    :param df: is the input dataframe of the donor or the recipient
    :param df_a: reference dataframe HLA A
    :param df_b: reference dataframe HLA B
    :param df_c: reference dataframe HLA C
    :param df_dr: reference dataframe HLA DRB1
    :param df_dq: reference dataframe HLA DQB1
    :param df_dp: reference dataframe HLA DPB1
    :param df_data: reference pandas.DataFrame with details about all the eplets (ep_data.csv)
    :param interlocus2: whether or not to take into account interlocus eplets for HLA of class II
    :param verifiedonly: whether or not to take into account only verified eplets

    :return: pd.DataFrame with eplets for each patient, for each locus
    """

    eplets_per_allele_dataframe: pd.DataFrame = df.map(
        lambda allele:
        _convert_to_eplets(
            allele,
            df_a,
            "ABC",
            df_data,
            interlocus2,
            verifiedonly
        )
        if allele[0] == "A"
        else (
            _convert_to_eplets(
                allele,
                df_b,
                "ABC",
                df_data,
                interlocus2,
                verifiedonly
            )
            if allele[0] == "B"
            else
            (
                _convert_to_eplets(
                    allele,
                    df_c,
                    "ABC",
                    df_data,
                    interlocus2,
                    verifiedonly
                )
                if allele[0] == "C"
                else
                (
                    _convert_to_eplets(
                        allele,
                        df_dr,
                        "DR",
                        df_data,
                        interlocus2,
                        verifiedonly
                    )
                    if allele[0:2] == "DR"
                    else
                    (
                        _convert_to_eplets(
                            allele,
                            df_dq,
                            "DQ",
                            df_data,
                            interlocus2,
                            verifiedonly
                        )
                        if allele[0:2] == "DQ"
                        else
                        (
                            _convert_to_eplets(
                                allele,
                                df_dp,
                                "DP",
                                df_data,
                                interlocus2,
                                verifiedonly
                            )
                            if allele[0:2] == "DP"
                            else
                            (
                                logging.error(f"Allele {allele} belongs to an undefined locus")
                            )
                        )
                    )
                )
            )
        )
    )
    # Keep this dataframe in case we would want the user to be able to hover over one of the alleles and to get the
    # eplet content of its allele.

    return eplets_per_allele_dataframe


def _extract_key_to_rank_eplets(eplet: str) -> int:
    """
    :param eplet: e.g. "9Y" or "26L"
    :return: the position but as a float so that sorting ain't gonna be done alphanumerically (9 or 26)

    If interlocus eplet, then return 1000 + position (e.g. 1037 for rqp37YA)
    """
    starts_with_number = re.search(r"^\d+", eplet)
    if starts_with_number:
        number = int(starts_with_number.group())
        return number
    else:
        extract_number = re.search(r"^.[PQR]*(\d+)", eplet)
        if extract_number:
            return 1000 + int(extract_number.group(1))  # 1000 will guarantee this comes last
        else:
            logging.error(f"Eplet {eplet} does not have the expected format")
            exit(55)


def _transform_eplet_load_detail(eplet_load_detail: pd.Series) -> pd.Series:
    """
    :param eplet_load_detail: pd.Series with the eplet load details
    :return: pd.Series with the eplet load details sorted and cleaned up
    """
    eplet_load_detail = eplet_load_detail.apply(list)
    eplet_load_detail = eplet_load_detail.apply(
        lambda list_: sorted(
            list_,
            key=_extract_key_to_rank_eplets
        )
    )
    eplet_load_detail = eplet_load_detail.astype(str)
    eplet_load_detail = (
        eplet_load_detail.replace("[]", "None")  # no regex no need to escape
        .replace("\\[", "", regex=True)
        .replace("\\]", "", regex=True)
        .replace("'", "", regex=True)
    )

    return eplet_load_detail


def _replace_null_alleles(df: pd.DataFrame) -> None:
    """
    :param df: pd.DataFrame with the typing details
    :return: None, the dataframe is modified inplace
    """
    # Replace Null Alleles
    df.replace(
        r'^(.*\*).*N$',
        r"\1",
        regex=True,
        inplace=True
    )
    df.replace(
        "DRB3*",
        "DRB345*",
        inplace=True
    )
    df.replace(
        "DRB4*",
        "DRB345*",
        inplace=True
    )
    df.replace(
        "DRB5*",
        "DRB345*",
        inplace=True
    )
    # Replace NaNs with ghost alleles
    for column in df.columns:
        df[column] = df[column].fillna(column[:-3] + "*")



import pandas as pd

from pelc.output_type import OutputType
from pelc.batch_eplet_comp import compute_epletic_load


def _is_valid_allele(allele: str) -> bool:
    """
    :param allele: allele to check
    :return: whether or not the allele is valid
    """
    return "*" in allele and ":" in allele


def _same_locus(allele1: str, allele2: str) -> bool:
    """
    :param allele1: First allele to compare
    :param allele2: Second allele to compare

    :return: whether or not the alleles are of the same locus
    """
    prefix_allele1 = allele1.split("*")[0]
    prefix_allele2 = allele2.split("*")[0]
    if prefix_allele1 == prefix_allele2:
        # A*01:01 and A*68:01 are of the same locus
        return True
    else:
        if "DRB" in prefix_allele1 and "DRB" in prefix_allele2:
            # DRB1*01:01 and DRB3*01:01 are of the same locus
            return True
        else:
            # A*01:01 and B*07:02 are not of the same locus
            return False


def simple_comparison(
        allele1: str,
        allele2: str,
        output_path: str | None,
        verifiedonly: bool = False,
        interlocus2: bool = True
) -> None | pd.DataFrame:
    """
    :param allele1: First allele to compare
    :param allele2: Second allele to compare
    :param output_path: Output path without the extension. If None, the output will be returned as a pandas.DataFrame.
    :param verifiedonly: How should the epletic charge be computed? Verified eplets only? Or all eplets?
    :param interlocus2: whether or not to take into account interlocus eplets (only relevant for HLA of class II)

    :return: None or a pandas.DataFrame with the results according to output_path
    """

    # Check if the alleles are valid
    if not _is_valid_allele(allele1) or not _is_valid_allele(allele2):
        raise ValueError("The alleles are not valid")
    else:
        # Check if the alleles are of the same locus
        if not _same_locus(allele1, allele2):
            raise ValueError("The alleles are not of the same locus")
        else:
            # Create a pandas DataFrame with the two alleles
            input_df_donor = pd.DataFrame(
                data={
                    "Donor": [allele1, allele2]
                },
                index=[
                    f"In {allele1} but not in {allele2}",
                    f"In {allele2} but not in {allele1}"
                ]
            )
            input_df_recipient = pd.DataFrame(
                data={
                    "Recipient": [allele2, allele1]
                },
                index=[
                    f"In {allele1} but not in {allele2}",
                    f"In {allele2} but not in {allele1}"
                ]
            )

            # Compute the epletic load
            return compute_epletic_load(  # type: ignore # we know that the output is a pandas.DataFrame or None
                input_df_donor,
                input_df_recipient,
                output_path,
                OutputType.DETAILS_AND_COUNT,
                verifiedonly=verifiedonly,
                interlocus2=interlocus2,
                simple_comparison=True
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pelc']

package_data = \
{'': ['*'], 'pelc': ['data/*']}

install_requires = \
['openpyxl>=3.0.10,<4.0.0',
 'pandas-stubs>=2.1.4.231227,<3.0.0.0',
 'pandas>=2.1.0,<3.0.0']

setup_kwargs = {
    'name': 'pelc',
    'version': '0.5.4.3',
    'description': 'Python Eplet Load Calculator',
    'long_description': '[![DOI](https://zenodo.org/badge/555576588.svg)](https://zenodo.org/badge/latestdoi/555576588)\n[![Downloads](https://pepy.tech/badge/pelc)](https://pepy.tech/project/pelc)\n# PELC (Python Eplet Load Calculator)\n\n### Overview\nPELC is a Python package designed to calculate efficiently the HLA Eplet Load (based on the\n[EpRegistry database](https://www.epregistry.com.br/)) between donors and recipients by loading in a pandas.DataFrame\nin `eplet_comparison.compute_epletic_load` the recipients\' and donors\' typings.  See minimal reproducible example for\nmore details.\n\n\n### Getting started\n#### Install from PyPI (recommended)\nTo use `pelc`, run `pip install pelc` in your terminal.\n\n\n#### Usage\n\n##### a. Comparing two alleles\nHere is a minimal example of how to use `pelc` to compare two alleles:\n```py\nfrom pelc.simple_comparison import simple_comparison\n\nsimple_comparison(\n    "A*68:01",\n    "A*68:02",\n    "output",  # file will be saved as output.csv in the current directory\n    verifiedonly=False,  # if True, only verified eplets will be considered, otherwise all eplets will be considered\n    interlocus2=True  # doesn\'t matter for class I alleles\n)\n```\nIn the `output.csv` file created in the current directory, you will find two rows: "In A&ast;68:02 but not in \nA&ast;68:01" and "In A&ast;68:01 but not in A&ast;68:02".\n\n##### b. Batch mode\nHere is a minimal example with the file [Template.xlsx](https://github.com/MICS-Lab/pelc/raw/main/Template.xlsx)\n(click to download):\n```py\nimport pandas as pd\n\nfrom pelc import batch_eplet_comp, batch_eplet_comp_aux, output_type\n\nif __name__ == "__main__":\n    input_path: str = "Template.xlsx"\n\n    output_path: str = "MyOutput"\n    input_df: pd.DataFrame = pd.read_excel(\n        input_path,\n        sheet_name="My Sheet",\n        index_col=0,\n    )\n\n    donordf: pd.DataFrame\n    recipientdf: pd.DataFrame\n    donordf, recipientdf = batch_eplet_comp_aux.split_dataframe(input_df)\n\n    batch_eplet_comp.compute_epletic_load(\n        donordf,\n        recipientdf,\n        output_path,\n        output_type.OutputType.DETAILS_AND_COUNT,\n        class_i=True,  # Compute class I eplets comparison?\n        class_ii=True,  # Compute class II eplets comparison?\n        verifiedonly=False,  # How should the epletic charge be computed? Verified eplets only? Or all eplets?\n        exclude=None,  # list of indices to exclude\n        interlocus2=True  # whether or not to take into account interlocus eplets for HLA of class II\n    )\n```\nNote that if a typing is unknown, one can use `A*`, `B*`, ..., `DPB1*` as the allele name for **both** recipients and\ndonors. If the allele is unknown for only of the two individuals, it is necessary to use `A*`, `B*`, ..., `DPB1*` for\nboth individuals otherwise the eplet mismatch computation will not be performed for this donor / recipient pair.\n\n#### Advanced usage:\n##### a. Not taking into account all loci (if they are not typed for example)\nIf one wants to determine the eplet mismatches between a donor and a recipient but without taking into account\na certain locus, one can use `A*`, `B*`, ..., `DPB1*` as the allele name for both recipients and donors on this locus\nand the eplet mismatch computation will only take into account the loci filled in.\n\n##### b. Not creating a file but generating a pandas.DataFrame\nIf one wants to generate a `pandas.DataFrame` directly, the `output_path` argument of `simple_comparison` can be \nset to `None`. The `pandas.DataFrame` will be returned by the function. Same goes for `compute_epletic_load`.\n\n\n#### Exit codes:\n```\n- 55: an eplet did not match the regular expression \'^\\d+\' (ABC, DR, DQ or DP) and it also did not match the regular\nexpression \'^.[PQR]*(\\d+)\' (interlocus2) either.\n```\n\n\n#### Unit tests\nTested on `Python 3.10.6` & `Python 3.11.1`.\n```\nplatform win32 -- Python 3.10.6, pytest-7.4.0, pluggy-1.2.0\nplugins: mypy-0.10.3\ncollected 39 items\n\nunit_tests_mypy.py ..                                                    [  5%]\nunit_tests_simple.py .                                                   [  7%]\npelc\\__init__.py .                                                       [ 10%]\npelc\\_input_sanity_check.py .                                            [ 12%]\npelc\\_open_epregistry_databases.py .                                     [ 15%]\npelc\\_unexpected_alleles.py .                                            [ 17%]\npelc\\batch_eplet_comp.py .                                               [ 20%]\npelc\\batch_eplet_comp_aux.py .                                           [ 23%]\npelc\\output_type.py .                                                    [ 25%]\npelc\\simple_comparison.py .                                              [ 28%]\ntests\\__init__.py .                                                      [ 30%]\ntests\\base_loading_for_tests.py .                                        [ 33%]\ntests\\test_eplet_mismatches.py .........                                 [ 56%]\ntests\\test_extract_key_to_rank_epletes.py ..                             [ 61%]\ntests\\test_is_valid_allele.py ..                                         [ 66%]\ntests\\test_only_one_chromosome.py ..                                     [ 71%]\ntests\\test_pelc.py ..                                                    [ 76%]\ntests\\test_same_locus.py ..                                              [ 82%]\ntests\\test_simple_comparison.py .....                                    [ 94%]\ntests\\test_unexpected_alleles.py ..                                      [100%]\n==================================== mypy =====================================\n\nSuccess: no issues found in 20 source files\n============================= 39 passed in 34.42s =============================\n```\n\n```\nplatform win32 -- Python 3.11.1, pytest-7.2.2, pluggy-1.0.0\nplugins: anyio-3.6.2, mypy-0.10.3\ncollected 39 items\n\nunit_tests_mypy.py ..                                                    [  5%]\nunit_tests_simple.py .                                                   [  7%]\npelc\\__init__.py .                                                       [ 10%]\npelc\\_input_sanity_check.py .                                            [ 12%]\npelc\\_open_epregistry_databases.py .                                     [ 15%]\npelc\\_unexpected_alleles.py .                                            [ 17%]\npelc\\batch_eplet_comp.py .                                               [ 20%]\npelc\\batch_eplet_comp_aux.py .                                           [ 23%]\npelc\\output_type.py .                                                    [ 25%]\npelc\\simple_comparison.py .                                              [ 28%]\ntests\\__init__.py .                                                      [ 30%]\ntests\\base_loading_for_tests.py .                                        [ 33%]\ntests\\test_eplet_mismatches.py .........                                 [ 56%]\ntests\\test_extract_key_to_rank_epletes.py ..                             [ 61%]\ntests\\test_is_valid_allele.py ..                                         [ 66%]\ntests\\test_only_one_chromosome.py ..                                     [ 71%]\ntests\\test_pelc.py ..                                                    [ 76%]\ntests\\test_same_locus.py ..                                              [ 82%]\ntests\\test_simple_comparison.py .....                                    [ 94%]\ntests\\test_unexpected_alleles.py ..                                      [100%]\n==================================== mypy =====================================\n\nSuccess: no issues found in 20 source files\n============================= 39 passed in 32.56s =============================\n```\n\n\n\n### About the source code\n- Follows [PEP8](https://peps.python.org/pep-0008/) Style Guidelines.\n- All functions are unit-tested with [pytest](https://docs.pytest.org/en/stable/).\n- All variables are correctly type-hinted, reviewed with [static type checker](https://mypy.readthedocs.io/en/stable/)\n`mypy`.\n- All functions are documented with [docstrings](https://www.python.org/dev/peps/pep-0257/).\n\n\n\n### Useful links:\n- [Corresponding GitHub repository](https://github.com/MICS-Lab/pelc)\n- [Corresponding PyPI page](https://pypi.org/project/pelc)\n\n\n\n### Citation\nIf you use this software, please cite it as below.\n\n- APA:\n```\nIf you use this software, please cite it as below. \n\nLhotte, R., Clichet, V., Usureau, C. & Taupin, J. (2022). \nPython Eplet Load Calculator (PELC) package (Version 0.5.4.3) [Computer software].\nhttps://doi.org/10.5281/zenodo.7254809\n```\n\n- BibTeX:\n```\n@software{lhotte_romain_2022_7526198,\n  author       = {Lhotte, Romain and\n                  Clichet, Valentin and\n                  Usureau, Cédric and\n                  Taupin, Jean-Luc},\n  title        = {Python Eplet Load Calculator},\n  month        = oct,\n  year         = 2022,\n  publisher    = {Zenodo},\n  version      = {0.5.4.3},\n  doi          = {10.5281/zenodo.7526198},\n}\n```\n',
    'author': 'JasonMendoza2008',
    'author_email': 'lhotteromain@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
